/* Mptn -- regexp-like pattern matcher suited for morphological analysis
 * Copyright (C) 2000  George Bronnikov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
  Paser for Mptns.

  $Id: mptn-parse.y 1.3 Sun, 23 Jan 2000 20:56:51 -0600 goga $
*/

%{
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include "mptn.h"
#include "mptn-chrset.h"
#include "mptn-private.h"

#define YYERROR_VERBOSE

enum {
  IDLEN = 32
};

enum syntype {
  ST_MPTN,
  ST_STR,
  ST_ANY,
  ST_ALT,
  ST_AND,
  ST_CAT
};

typedef struct synval {
  enum syntype type;
  union {
    mptn_t *mptn;
    chr *str;
    GPtrArray *arr;
  } u;
} synval_t;;

/* The type of the argument to pass to yyparse and yylex */
typedef struct ly_param {
  mptn_t *res;		/* The result of calling yyparse */
  chr *str;		/* The string to parse from */
  GArray *varnames;
  mptn_control_t *mptn_control;
} ly_t;
  
static chr yy_readchr (ly_t *lyp);
static void yy_normalize (ly_t *lyp,
			  synval_t *elem, enum syntype reqtype);

static void yy_binop (ly_t *lyp,
		      synval_t *f1, synval_t *f2, synval_t *res, enum syntype op);
static void yy_rep  (ly_t *lyp,
		     synval_t *arg, synval_t *res,
		     gboolean empty_allowed, gboolean rep_allowed);
static int yyerror (char* str);
/* static int yylex (YYSTYPE* lvalp, void *param); -- YYSTYPE bot yet defined here -- Goga */

#define YYPARSE_PARAM param
#define YYLEX_PARAM param

%}

%union {
  synval_t v;
}

%token <v> CHR CHRSET VARIABLE ANY

%left 		'&' '|'
%left 		CAT
%nonassoc  	'+' '*' '?' '>' '<'

%type <v> mptn_start mptn mptn_nonempty

%start mptn_start
%pure_parser

%%

mptn_start: mptn 
	     {
	       yy_normalize ((ly_t*)param, &$1, -1);

	       /* Set the varnames for the toplevel mptn */
               $1.u.mptn->varnames = ((ly_t*)param)->varnames;

	       /* Fill the Res field of the yyparse parameter */
	       ((ly_t*)param)->res = $1.u.mptn;
	     }

mptn_nonempty:
      mptn '&' mptn 
        {
	  yy_binop ((ly_t*)param, &$1, &$3, &$$, ST_AND);
	}
    | mptn '|' mptn 
        {
	  yy_binop ((ly_t*)param, &$1, &$3, &$$, ST_ALT);
	}
    | mptn_nonempty '*' 
        {
	  yy_rep ((ly_t *)param, &$1, &$$, TRUE, TRUE);
	}
    | mptn_nonempty '+'
        {
	  yy_rep ((ly_t *)param, &$1, &$$, FALSE, TRUE);
	}
    | mptn_nonempty '?'
	{
	  yy_rep ((ly_t *)param, &$1, &$$, TRUE, FALSE);
	}
    | mptn_nonempty mptn_nonempty  %prec CAT
	{
	  if ($1.type == ST_STR
	      && $2.type == ST_STR)	/* Concatenating two strings */
	    {
	      $$.type = ST_STR;
	      $$.u.str = g_new (chr, chra_len ($1.u.str) + chra_len ($2.u.str) + 1);
	      chra_copy ($$.u.str, $1.u.str);
	      chra_cat  ($$.u.str, $2.u.str);
	      
	      g_free ($1.u.str);
	      g_free ($2.u.str);
	    }
	  else
	    yy_binop ((ly_t*) param, &$1, &$2, &$$, ST_CAT);
	}
    | mptn_nonempty '>'
	{
	  yy_normalize ((ly_t*) param, &$1, ST_CAT);

	  if ($1.type == ST_MPTN)
	    {
	      $1.u.mptn->flags = MPTN_LONGER;
	    }
	  else
	    {
	      g_assert ($1.type == ST_CAT);

	      ((mptn_t*) g_ptr_array_index ($1.u.arr, $1.u.arr->len - 1))->flags = MPTN_LONGER;
	    }
	  $$ = $1;
	}
    | mptn_nonempty '<'
	{
	  yy_normalize ((ly_t *)param, &$1, ST_CAT);

	  if ($1.type == ST_MPTN)
	    {
	      $1.u.mptn->flags = MPTN_SHORTER;
	    }
	  else
	    {
	      g_assert ($1.type == ST_CAT);

	      ((mptn_t*) g_ptr_array_index ($1.u.arr, $1.u.arr->len - 1))->flags = MPTN_SHORTER;
	    }
	  $$ = $1;
	}
    |  CHR
        {
	  $$ = $1;
	}
    | CHRSET
	{
	  $$ = $1;
	}
    | VARIABLE
	{
	  $$ = $1;
	}
    | ANY
	{
	  $$ = $1;
	}
    | '(' mptn ')'
	{
	  $$ = $2;
	}
    | '`' mptn '\''
	{
	  yy_normalize ((ly_t *)param, &$2, -1);
	  $$.type = ST_MPTN;
	  $$.u.mptn = mptn_ta_matcher_init (((ly_t*)param)->mptn_control,
					    $2.u.mptn);
	}
    ;

/* 
   Push the empty variant to the end of the grammar file, so that it is never default.
   How do I deal with this in a more legal way? -- Goga
*/
mptn: mptn_nonempty 
	{$$ = $1;}
    | /* empty */
	{
	  $$.type = ST_MPTN;
	  $$.u.mptn = mptn_string_matcher_init (((ly_t*)param)->mptn_control, 	/* control */
						"", NULL,	/* begin, end */
						NULL,		/* varnames -- not used there */
						NULL);		/* data */
	}
  
%%

/* Lexical analyzer */
int
yylex (YYSTYPE *lvalp, void *param)
{
  chr cc;
  ly_t *lyp = param;

  switch ((cc = yy_readchr(lyp)))
    {
    case 0:
      return 0;
    case '\\':
      {
	chr *str;

	cc = yy_readchr (param);
	str = g_malloc (2 * sizeof (chr));
	str[0] = cc;
	str[1] = '\0';

	lvalp->v.type = ST_STR;
	lvalp->v.u.str = str;
      }
      return CHR;
    case '[':
      {
	gint c;
	gboolean rev = FALSE;
	mptn_chrset_t *chrset;
	
	chrset = mptn_chrset_create ();
      
	if ((c = yy_readchr (lyp)) == '^')
	  {
	    rev = TRUE;
	    c = yy_readchr (lyp);
	  }

	for (;;)
	  {
	    switch (c)
	      {
	      case EOF:
		g_warning ("Unfinished charset constant in mptn");
		return 0;
	      case ']':
		if (rev)
		  mptn_chrset_reverse (chrset);
		lvalp->v.type = ST_MPTN;
		lvalp->v.u.mptn = mptn_mptn_chrset_init (lyp->mptn_control, chrset);
		mptn_chrset_destroy (chrset);
		return CHRSET;
	      case '\\':
		mptn_chrset_add (chrset, yy_readchr (lyp));
		break;
	      default:
		mptn_chrset_add (chrset, c);
		break;
	      }
	    c = yy_readchr (lyp);
	  }
	/* Should never get here */
	g_assert (FALSE);
      }
    case '&':
    case '|':
    case '+':
    case '*':
    case '?':
    case '(':
    case ')':
    case '`':
    case '\'':
    case '>':
    case '<':
      return cc;
    case '.':
      {
	lvalp->v.type = ST_ANY;
	return ANY;
      }
    case '{':
      {
	GString * gstr;
	chr c;
	chr *str;
	int i;
	i = 0;
	gstr = g_string_new ("");
	while (isalnum(c = yy_readchr(lyp)) || c == '_')
	  g_string_append_c (gstr, c);

	str = gstr->str;
	g_string_free (gstr, FALSE);	/* Don't free segment */

	if (c == '}')	/* Plain variable */
	  {
	    lvalp->v.type = ST_MPTN;
	    lvalp->v.u.mptn = mptn_var_matcher_init (lyp->mptn_control,	/* control */
						     str, NULL,		/* begin, end */
						     lyp->varnames, /* varnames */
						     NULL);		/* data */
	    g_free (str);

	    return VARIABLE;
	  }
	else if (c == ':')	/* A matcher with arguments */
	  {
	    GString *gs_arg;
	    chr *s_arg;
	    int bcou;

	    bcou = 0;
	    
	    gs_arg = g_string_new ("");
	    
	    while ((c = yy_readchr (lyp)))		/* Time to use Flex? -- Goga */
	      {
		switch (c)
		  {
		  case '{':
		  case '[':
		  case '(':
		    bcou++;
		    break;

		  case '}':
		    if (bcou == 0) /* end of arguments */
		      goto ENDARG;
		    
		    /* fall through */
		  case ']':
		  case ')':
		    bcou --;
		    if (bcou < 0)
		      {
			g_warning ("Bad arguments for matcher {%s:%s%c", str, gs_arg->str, c);
			return 0;
		      }
		    break;
		    
		  case '\\':	/* escape chr */
		    c = yy_readchr (lyp);	/* will miss it */
		    g_string_append_c (gs_arg, c);
		    break;

		  default:
		    break;
		  } /* switch */

		g_string_append_c (gs_arg, c);
	      }		/* arg reading loop */

	    /* Reading arguments, got to the end of the input */
	    g_warning ("Unexpected end of input %s:%s", str, gs_arg->str);
	    return 0;
	    
	  ENDARG:			 /* Found the ending '}' */
	    s_arg = gs_arg->str;
	    g_string_free (gs_arg, FALSE);

	    /* Here: str is the name of the matcher, s_arg contains its arguments */
	    {
	      GQuark scode;
	      mptn_matcher_ops_t *ops;
	      gpointer param;

	      scode = chra_quark_from_string (str);
	      ops = mptn_control_get_matcher (lyp->mptn_control, scode);
	      if (!ops)
		{
		  g_warning ("Unknown matcher: %s", str);
		  return 0;
		}

	      param = mptn_control_get_matcher_param (lyp->mptn_control, scode);

	      lvalp->v.type = ST_MPTN;
	      lvalp->v.u.mptn = (*ops->init) (lyp->mptn_control,	/* control */
					      s_arg, NULL, 	/* begin, end */
					      lyp->varnames, /* varnames */
					      param);		/* data */
	      lvalp->v.u.mptn->ops = ops; 	/* Should I do this by force? -- Goga */

	      g_free (str);
	      g_free (s_arg);
	      
	      return VARIABLE;
	    }
	  }
	else
	  {
	    g_warning ("Mptn: bad identifier: %s%c", str, c);
	    return 0;
	  }
      }
    default:
      {
	char *str;

	str = g_malloc (2 * sizeof (chr));
	str[0] = cc;
	str[1] = '\0';
	
	lvalp->v.type = ST_STR;
	lvalp->v.u.str = str;
	return CHR;
      }
    }
}

/* 
   Normalize an expression (of YYSTYPE) to ST_MPTN, 
   unless the type is Reqtype.
   Works in place.
*/

/*
  Initialization functions depending on expr type
*/
static struct {
  enum syntype type;
  mptn_t *(*fun) (mptn_control_t *control, mptn_t **subs);
} initfuns[] = {
  {ST_ALT, mptn_alt_matcher_init},
  {ST_CAT, mptn_cat_matcher_init},
  {ST_AND, mptn_conj_matcher_init},
  {-1, NULL}
};

static void
yy_normalize (ly_t *lyp, synval_t *elem, enum syntype reqtype)
{
  switch (elem->type)
    {
    case ST_STR:
      {
	chr *str = elem->u.str;
	elem->type = ST_MPTN;
	elem->u.mptn = mptn_string_matcher_init (lyp->mptn_control, 	/* control */
						 str, NULL,	/* begin, end */
						 lyp->varnames,	/* varnames */
						 NULL);		/* data */
	g_free (str);
      }
      break;
    case ST_ANY:
      {
	elem->type = ST_MPTN;
	elem->u.mptn = mptn_any_matcher_init (lyp->mptn_control,/* control */
					      FALSE, FALSE);	/* zero, repetition */
      }
      break;
    case ST_MPTN:
      break;	/* nothing to do */
    case ST_ALT:
    case ST_AND:
    case ST_CAT:
      if (elem->type != reqtype)
	{
	  mptn_t **subs;
	  mptn_t **tt;
	  int i;
	  
	  g_ptr_array_add (elem->u.arr, NULL); /* end the alts array with a zero pointer */
	  subs = (mptn_t**) elem->u.arr->pdata;
	  g_ptr_array_free (elem->u.arr, FALSE);
	  
	  /* Find the init function */
	  for (i = 0; initfuns[i].type != -1; i++)
	    if (initfuns[i].type == elem->type)
	      break;
	  
	  g_assert (initfuns[i].type != -1); 	/* we should break ut of the loop somewhere */
	  
	  elem->type = ST_MPTN;
	  elem->u.mptn = (*initfuns[i].fun) (lyp->mptn_control, subs);
	  
	  /* Unref the alternatives */
	  for (tt = subs; *tt; tt++)
	    mptn_free (*tt);
	  
	  g_free (subs);
	}
      break;
    default:
      g_error ("Unknown syntax element type: %d", elem->type);
      break;
    }
}

/* Most binary operators are treated by this procedure */
/* F1, F2 are the operands; Res is the result */
static void 
yy_binop (ly_t *lyp, synval_t *f1, synval_t *f2, synval_t *res, enum syntype op)
{
  yy_normalize (lyp, f1, op);	/* Convert both operands to Mptns */
  yy_normalize (lyp, f2, op);

  /* 
     Now check if one of the operands is already 
     formed by an operation of the needed type 
  */
  if (f1->type == op && f2->type == op) 	/* Both are */
    {
      int i;
      
      /* Add the elements of f2 to the end of f1 */
      for (i = 0; i < f2->u.arr->len; i++)
	g_ptr_array_add (f1->u.arr, g_ptr_array_index (f2->u.arr, i));

      g_ptr_array_free (f2->u.arr, TRUE);
      
      /* All the refcounts remain the same */

      *res = *f1;
    }
  else if (f1->type == op)
    {
      g_ptr_array_add (f1->u.arr, f2->u.mptn);
      *res = *f1;
    }
  else if (f2->type == op)
    {
      /* GPtrArray has no prepend operation, have to copy the entire array */
      
      int i;
      
      res->type = op;
      res->u.arr = g_ptr_array_new ();
      g_ptr_array_add (res->u.arr, f1->u.mptn);

      for (i = 0; i < f2->u.arr->len; i++)
	g_ptr_array_add (res->u.arr, g_ptr_array_index (f2->u.arr, i));

      g_ptr_array_free (f2->u.arr, FALSE);

      /* All the refcounts remain the same */
    }
  else
    {
      res->type = op;
      res->u.arr = g_ptr_array_new ();
      g_ptr_array_add (res->u.arr, f1->u.mptn);
      g_ptr_array_add (res->u.arr, f2->u.mptn);

      /* Refcounts of mptns remain the same */
    }
}

/* A repetition-like operation (*, +, ?) */
static void 
yy_rep  (ly_t *lyp,
	 synval_t *arg, synval_t *res, gboolean empty_allowed, gboolean rep_allowed)
{
  if (arg->type == ST_ANY)
    {
      /* A special case: ".*", ".+", ".?" */
      res->type = ST_MPTN;
      res->u.mptn = mptn_any_matcher_init (lyp->mptn_control, empty_allowed, rep_allowed);
    }
  else
    {
      yy_normalize (lyp, arg, -1);
      
      res->type = ST_MPTN;
      res->u.mptn = mptn_rep_matcher_init (lyp->mptn_control, arg->u.mptn, 
					   empty_allowed, rep_allowed);	/* zero, rep */
      
      mptn_free (arg->u.mptn);	/* Decrease the refcount */
    }
}


/* Error reporting */
static int
yyerror (char *str)
{
  mptn_parse_error ("Mptn parsing error: %s", str);
  return 0;
}

/* Read the next character (or 0 if the string is over) */
static chr
yy_readchr (ly_t *lyp)
{
  if (*lyp->str)
    return *(lyp->str++);
  else
    return (chr) 0;
}

/*
  Parse an expression -- this is the external entry into this file.
*/
mptn_t *
mptn_parse (mptn_control_t *control, chr *str)
{
  extern int yyparse (void *param);
  
  ly_t ly;

  ly.str = str;
  ly.varnames = g_array_new (FALSE, TRUE, sizeof (GQuark));
  ly.mptn_control = control;

  if (yyparse(&ly))
    g_error ("Error parsing mptn: %s", str);

  return ly.res;
}
