/* Mptn -- regexp-like pattern matcher suited for morphological analysis
 * Copyright (C) 2000  George Bronnikov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
  Operations on Mptns -- compiled patterns.

  $Id: mptn.c 1.3 Sun, 23 Jan 2000 20:56:51 -0600 goga $
*/

#include "mptn.h"
#include "mptn-chrset.h"
#include "mptn-private.h"

/* 
   An internal function to start the matching iterator,
   may be called recursively.
*/
static mptn_ectl_t *
mptn_exec_start_internal (mptn_t *mptn, chr *begin, chr *end,
			  mptn_var_t *vars, 		/* Where to store variables */
			  guint flags,			/* so far, only VARS_ALLOCATED */
			  guint stage,			/* At which stage we are */
			  gpointer data);

/*
  Allocation and basic setup.
*/
mptn_t *
mptn_new (mptn_control_t *control)
{
  mptn_t *res;

  res = g_malloc0 (sizeof (mptn_t));
  res->magic = MPTN_MAGIC;
  res->refcnt = 1;
  res->control = mptn_control_normalize (control);
  res->varnames = NULL;

  return res;
}

/*
  Free an mptn.
*/
void
mptn_free (mptn_t *mptn)
{
  if (!mptn) return;

  MPTN_CHECK (mptn);

  if (--mptn->refcnt <= 0)
    {
      if (mptn->ops->destroy)
	mptn->ops->destroy (mptn->mctl);
      if (mptn->varnames)
	g_array_free (mptn->varnames, TRUE);
      g_free (mptn);
    }
}

/*
  Dump an mptn.
*/
void 
mptn_print (mptn_t *mptn, FILE *fp, guint offset)
{
  MPTN_CHECK (mptn);
  g_return_if_fail (fp != NULL);

  if (mptn->ops->print == NULL)
    {
      int i;

      for (i = 0; i < offset; i++)
	fputc (' ', fp);

      fputs ( "<UNPRINTABLE>\n", fp);
    }
  else
    (*mptn->ops->print) (mptn, fp, offset);
}

/*
  Increase the reference count of an mptn.
*/
void
mptn_refinc (mptn_t *mptn)
{
  MPTN_CHECK (mptn);
  mptn->refcnt++;
}

/*
  Get the index of a given variable inside the variable names' array.
  If there is no such variable yet, a place for a new variable is created
  (this should only be done during matcher registration).
  The name is given in the form of a GQuark.

  We don't expect this function to be called too frequently or to have long
  variable lists, thus linear search.
*/
guint
mptn_var_get_index (GArray *arr, GQuark vcode)
{
  int i;

  g_return_val_if_fail (vcode != 0, -1);

  for (i = 0; i < arr->len; i++)
    if (g_array_index (arr, GQuark, i) == vcode)
      return i;

  /* Not found, need to add */
  g_array_append_val (arr, vcode);

  return arr->len - 1;
}


/* ----  Applying a pattern */

/* 
   Start matching an mptn against a string
   Returns a cotrol structure.
*/
mptn_ectl_t *
mptn_exec_start (mptn_t *mptn, chr *begin, chr *end, gpointer data)
{
  mptn_var_t *vars;
  int i;

  g_return_val_if_fail (begin != NULL, NULL);
  MPTN_CHECK (mptn);

  if (!end)
    end = begin + chra_len (begin);

  /*  Try to drop out as soon as possible */
  if (end - begin < mptn->min_len
      || end - begin > mptn->max_len)
    return NULL;

  if (mptn->varnames)
    {
      vars = g_malloc0 (sizeof (mptn_var_t) * (mptn->varnames->len + 1)); 
      
      /* Set the names of the variables */
      for (i = 0; i < mptn->varnames->len; i++)
	vars[i].vcode = g_array_index (mptn->varnames, GQuark, i);

      vars[mptn->varnames->len].vcode = 0;
    }
  else
    vars = NULL;
      
  return mptn_exec_start_internal (mptn, begin, end,
				   vars, MPTN_ECTL_VARS_ALLOCATED, 1,
				   data);
}

/*
  A variant of Mptn_ectl_start when the array for the variables
  is provided by the caller (and may be partially filled)
*/
mptn_ectl_t *
mptn_exec_start_with_vars (mptn_t *mptn, chr *begin, chr *end,
			   mptn_var_t *vars, guint stage,
			   gpointer data)
{
  g_return_val_if_fail (begin != NULL, NULL);
  MPTN_CHECK (mptn);

  if (!end)
    end = begin + chra_len (begin);

  /*  Try to drop out as soon as possible */
  if (end - begin < mptn->min_len
      || end - begin > mptn->max_len)
    return NULL;

  return mptn_exec_start_internal (mptn, begin, end,
				   vars, 0, stage,
				   data);
}

/* 
   An internal function to start the matching iterator,
   may be called recursively.
*/
static mptn_ectl_t *
mptn_exec_start_internal (mptn_t *mptn, chr *begin, chr *end,
			  mptn_var_t *vars, 		/* Where to store variables */
			  guint flags,			/* so far, only VARS_ALLOCATED */
			  guint stage,			/* At which stage we are */
			  gpointer data)
{
  mptn_ectl_t *res;

  MPTN_CHECK (mptn);
  g_assert (begin != NULL);
  g_assert (end != NULL);

  mptn_vars_clear (vars, stage);

  res = g_malloc (sizeof (mptn_ectl_t));
  res->magic = MPTN_ECTL_MAGIC;
  res->flags = flags;
  res->mptn = mptn;
  res->vars = vars;
  res->stage = stage;
  res->ictl = mptn->ops->iter_start (mptn, begin, end,
				     vars, stage,
				     data);

  if (!res->ictl) 	/* Refuses to match */
    {
      mptn_ectl_free (res);
      return NULL;
    }

  return res;
}

/*
  Get the next match.
*/
mptn_ectl_t *
mptn_exec_step (mptn_ectl_t *ectl)
{
  MPTN_ECTL_CHECK (ectl);

  if (ectl->mptn->ops->iter_step)
    {
      ectl->ictl = ectl->mptn->ops->iter_step (ectl->ictl);
    }
  else		
    ectl->ictl = NULL; 	/* No step => just one match */

  if (!ectl->ictl) 	/* Refuses to match */
    {
      mptn_ectl_free (ectl);
      return NULL;
    }

  return ectl;
}

/*
  Get the variables' values from a Ectl.
*/
mptn_var_t *
mptn_ectl_vars (mptn_ectl_t *ectl)
{
  MPTN_ECTL_CHECK (ectl);
  return ectl->vars;
}

/*
  Get the current matching stage from a Ectl.
*/
guint 
mptn_ectl_stage (mptn_ectl_t *ectl)
{
  guint res = 0;

  MPTN_ECTL_CHECK (ectl);

  if (ectl->mptn->ops->stage)	/* The matcher does mess with stages */
    res =  ectl->mptn->ops->stage (ectl->ictl);

  return res ? res : ectl->stage;
}

/*
  If the matching cycle is finished prematurely,
  you need to destroy Ectl.
*/
void 
mptn_ectl_free (mptn_ectl_t *ectl)
{
  MPTN_ECTL_CHECK (ectl);

  if (ectl->mptn->ops->ictl_free && ectl->ictl)
    ectl->mptn->ops->ictl_free (ectl->ictl);

  mptn_vars_clear (ectl->vars, ectl->stage);
  if (ectl->flags & MPTN_ECTL_VARS_ALLOCATED)
    g_free (ectl->vars);

  g_free (ectl);
}

/*
  Get just one match, filling Vars, whinch has Nvars elements.
  Returns: whether the match was successful.
*/
gboolean
mptn_exec_one (mptn_t *mptn, chr* begin, chr *end,
	       int nvars, mptn_var_t *vars,
	       gpointer data)
{
  mptn_ectl_t *ectl;

  g_return_val_if_fail (begin != NULL, FALSE);
  MPTN_CHECK (mptn);
  if (!end)
    end = begin + chra_len (begin);
  
  ectl = mptn_exec_start (mptn, begin, end, data);
  if (ectl) 	/* Success */
    {
      if (vars)
	memcpy (vars, mptn_ectl_vars (ectl), nvars * sizeof (mptn_var_t));

      mptn_ectl_free (ectl);
      return TRUE;
    }
  else		/* Failure */
    return FALSE;
}

/*
  Substitute the variables from Vars into the pattern described by Mptn;
  return a string allocated by g_malloc or NULL on failure.
*/
chr *
mptn_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data)
{
  if (!mptn->ops->subst)
    return NULL;

  return (*mptn->ops->subst) (mptn, vars, data);
}

/* Clear the values of the variables younger than Stage */
void
mptn_vars_clear (mptn_var_t *vars, guint stage)
{
  if (!vars)	/* There may be no vars */
    return;

  /* Just scan the variables linearly */
  for (; vars->vcode; vars++)
    if (vars->stage >= stage)
      {
	if ((*vars).allocated)
	  g_free ((*vars).beg);
	(*vars).allocated = FALSE;
	(*vars).beg = (*vars).end = NULL;
	(*vars).stage = 0;
      }
}

